#include "page_queue_shared.h"



/* Discard all queued flexpages. */

void PageQueueShared::close(Memory *memory)
{
    discard(_queue, memory);
}

/* Keep waiting for a potential queue non-empty condition.
   Then, attempt to pop an entry from the queue. */

void PageQueueShared::pop(PageOwner **owner, Flexpage **flexpage)
{
    std::unique_lock<std::mutex> guard(_lock);
    QueueEntry entry;

    while (1)
    {
        if (_pop(&entry))
        {
            *owner = entry.owner;
            *flexpage = entry.flexpage;
            return;
        }
        else
            _counter.wait(guard);
    }
}

/* Check the queue for entries, returning false if no entries are available,
   returning true and providing the details if an entry can be removed from the
   front of the queue. */

bool PageQueueShared::_pop(QueueEntry *entry)
{
    if (_queue.empty())
        return false;

    *entry = _queue.front();
    _queue.pop_front();

    /* Remove any position reference for the flexpage. */

    Positions::iterator position = _positions.find(entry->flexpage);

    if (position != _positions.end())
        _positions.erase(position);

    return true;
}

/* Push an entry for the given owner and flexpage to the queue. */

void PageQueueShared::push(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    /* Record the entry and a position reference for the flexpage. */

    _queue.push_back((QueueEntry) {flexpage, owner});

    Queue::iterator last = _queue.end();
    last--;
    _positions.insert(Position(flexpage, last));

    _counter.notify_one();
}

/* Push an entry to the front of the queue. */

void PageQueueShared::push_front(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    _queue.push_back((QueueEntry) {flexpage, owner});
    _positions.insert(Position(flexpage, _queue.begin()));

    _counter.notify_one();
}

/* Remove an entry for the given owner and flexpage from the queue. */

bool PageQueueShared::remove(PageOwner *owner, Flexpage *flexpage)
{
    return PageQueue::remove(_queue, _positions, owner, flexpage);
}

// vim: tabstop=4 expandtab shiftwidth=4
